/* Zplet, a Z-Machine interpreter in Java */
/* Copyright 1996 Matthew T. Russotto */
/* All Rights Reserved */
/* You may freely copy, modify, and make derivative works from this code
   provided my copyright notice is retained and any modifications are
   clearly marked */

import java.util.*;
import java.awt.*;
import java.net.*;
import java.io.*;
import java.applet.Applet;

//Race condition save vector class
//Thanks to werpu@inflab.uni-linz.ac.at

class NoSuchKeyException extends Exception {
	public NoSuchKeyException() {
		super();
	}
	
	public NoSuchKeyException(String s) {
		super(s);
	}
}
class SyncVector extends Vector {
	public SyncVector() {
		super();
	}

	public synchronized Object syncPopFirstElement() {
		Object first = syncFirstElement();
		if (first != null)
				removeElementAt(0);
		return first;
	}
	
	public synchronized Object syncFirstElement() {
		Object first = null;
		try {
			first = super.firstElement();
		} catch (NoSuchElementException booga) {}
		try {
			if (first==null) wait();
			else return first;
		} catch (InterruptedException booga) {}
		return null;
	}

	public synchronized void syncAddElement(Object obj) {
		super.addElement(obj);
		notify();
	}
}

 class ZWindow {
		 final static int ROMAN = 0;
		 final static int REVERSE = 1;
		 final static int BOLD = 2;
		 final static int ITALIC = 4;
		 final static int FIXED = 8;

		 final static char FIRST_STYLE = '\u8000';
		 final static char BUF_ROMAN = (char)(FIRST_STYLE + ROMAN);
		 final static char BUF_REVERSE = (char)(FIRST_STYLE + REVERSE);
		 final static char BUF_BOLD = (char)(FIRST_STYLE + BOLD);
		 final static char BUF_ITALIC = (char)(FIRST_STYLE + ITALIC);
		 final static char BUF_FIXED = (char)(FIRST_STYLE + FIXED);
		 final static char LAST_STYLE = '\u800f';
		
		 final static int NORMAL_FONT = 1;
		 final static int PICTURE_FONT = 2;
		 final static int GRAPHICS_FONT = 3;
		 final static int FIXED_FONT = 4;

		 final static char FIRST_FONT = '\u8010';
		 final static char BUF_NORMAL_FONT = '\u8010';
		 final static char BUF_PICTURE_FONT = '\u8011';
		 final static char BUF_GRAPHICS_FONT = '\u8012';
		 final static char BUF_FIXED_FONT = '\u8013';
		 final static char LAST_FONT = '\u8013';

		 ZScreen myscreen;
		 int top, left, width, height;
		 int cursorx, cursory;
		 boolean buffer, wrap, scroll, transcriptmode;
		 int curzfont = NORMAL_FONT;
		 int curzstyle = ROMAN;
		 Font curfont;
		 String linebuffer;
		 int line_counter;
		 int zforeground, zbackground;
		 int residual;

		 public ZWindow(ZScreen screen) {
				 top = 0;
				 left = 0;
				 width = 10;
				 height = 10;
				 cursorx = 0;
				 cursory = 0;
				 line_counter = 0;
				 residual = 0;
				 myscreen = screen;
				 curfont = screen.fixedfont;
				 buffer = true;
				 wrap = true;
				 scroll = true;
				 transcriptmode = true;
				 linebuffer = "";
				 zforeground = screen.zforeground;
				 zbackground = screen.zbackground;
		 }

		 void reset_line_count() {
				 line_counter = 0;
		 }

		 void count_line() {
				 line_counter++;
		 }
		
		 void check_for_more() {
				 String more  = "[MORE]";
				 String blank = "         ";

				 if (line_counter >= (height-1)) {
						 myscreen.settext(top + cursory, left, more.toCharArray(), 0, more.length(), true, myscreen.fixedfont);
						 myscreen.read_code();
						 myscreen.settext(top + cursory, left, blank.toCharArray(), 0, blank.length(), false, myscreen.fixedfont);
						 line_counter = 0;
				 }
				
		 }
		
		 void erase_line(short arg) {
				char spaces[] = new char[width];
				short i;
				
				for (i = 0; i < width; i++)
					spaces[i] = ' ';
				if (arg == 1)
					myscreen.settext(cursory, left + cursorx, spaces, 0, width - cursorx);
		 }
		
		 public void setbuffermode(boolean buffermode)
		 {
				 if (buffer && !buffermode) {
						 flush();
				 }
				 buffer = buffermode;
		 }

		 public void setwrapmode(boolean wrapmode)
		 {
				 if (buffer) {
						 flush();
				 }
				 wrap = wrapmode;
		 }

		 public void set_transcripting(boolean transcriptmode)
		 {
				 this.transcriptmode = transcriptmode;
		 }

		 public boolean transcripting()
		 {
				return transcriptmode;
		 }

		 public void setscroll(boolean newscroll)
		 {
				 scroll = newscroll;
		 }

		 public void moveto(int newleft, int newtop)
		 {
				 left = newleft;
				 top = newtop;
		 }

		 public void resize (int newwidth, int newheight)
		 {
				 width = newwidth;
				 height = newheight;
				 if ((cursorx >= width) || (cursory >= height)) {
						 movecursor_noflush(0,0);
				 }
		 }

		 public int getLeft() {
				 return left;
		 }

		 public int getTop() {
				 return top;
		 }

		 public int getWidth() {
				 return width;
		 }

		 public int getHeight() {
				 return height;
		 }

		 public int getlines() {
				 return height;
		 }

		 public int getchars() {
				 return width;
		 }

		 public int getx() {
				 return cursorx;
		 }

		 public int gety() {
				 return cursory;
		 }

		 void movecursor_noflush(int x, int y)
		 {
				 cursorx = x;
				 cursory = y;
		 }

		 public void movecursor(int x, int y)
		 {
				 flush();
				 cursorx = x;
				 cursory = y;
		 }
		 
		 public void printzascii(short ascii) {
				 short zascii[] = new short[1];

				 zascii[0] = ascii;
				 printzascii(zascii);
		 }

		 public void printzascii(short ascii[]) {
				 /* short because the Z-machine can put out 10-bit codes */
				 char unicode[];
				 int i;
				
				 unicode = new char[ascii.length];

				 for (i = 0; i < unicode.length; i++) {
						 unicode[i] = ZScreen.zascii_to_unicode(ascii[i]);
				 }

				 if (buffer) {
						 bufferchars(unicode);
				 }
				 else {
						 drawchars(unicode,0, unicode.length);
				 }
		 }

		 public void flush() {
		 		residual = charsWidth(linebuffer.toCharArray(), 0, linebuffer.length());
				drawstring(linebuffer);
				linebuffer = "";
//				System.err.println();
//				System.err.println("linebuffer.length() = " + linebuffer.length());
		 }
 
		 public void newline() {
				 newline(true);
		 }

		 protected void newline(boolean flushbuffer) {
				 if (myscreen.zforeground != zforeground)
						 myscreen.setZForeground(zforeground);
				 if (myscreen.zbackground != zbackground)
						 myscreen.setZBackground(zbackground);

				 if (flushbuffer)
						 flush();
						 
				 residual = 0;

				 if (cursory == height-1) {
						 if (scroll)
								 myscreen.scrollLines(top, height, 1);
						 movecursor_noflush(0, cursory);
				 }
				 else {
						 movecursor_noflush(0, cursory+1);
				 }
				 count_line();
//				 System.err.println();
		 }

		 public synchronized void bufferchars(char chars[]) {
				 int last;
				 int space;
				 String printstring;

//				 linebuffer = linebuffer + chars;
				 linebuffer = new StringBuffer(linebuffer).append(chars).toString();

				 if (wrap) {
						 last = linebuffer.length();
						 while ((residual + charsWidth(linebuffer.toCharArray(), 0, last))
								 > (myscreen.size().width)) {
								 space = linebuffer.lastIndexOf(' ', last);
								 if (space == -1) {
										 while ((residual + charsWidth(linebuffer.toCharArray(), 0, last)) > 
														(width * myscreen.charwidth()))
												 last--;
										 drawchars(linebuffer.toCharArray(), 0, last);
										 linebuffer = linebuffer.substring(last);
										 newline(false);
										 last = linebuffer.length();
								 }
								 else if ((residual + charsWidth(linebuffer.toCharArray(), 0, space)) <=
												  (myscreen.size().width)) {
										 printstring = linebuffer.substring(0, space);
										 drawstring(printstring);
										 while ((space < linebuffer.length())
												 && (linebuffer.charAt(space) == ' ')) {
												 space++;
										 }
										 linebuffer = linebuffer.substring(space);
										 last = linebuffer.length();
										 newline(false);
								 }
								 else
										 last = space-1;
						 }
				}
		 }

		 public void clear() {
				 int i;
				 char spaces[] = new char[width];

				 for (i = 0; i < width; i++)
						 spaces[i] = ' ';
				
				 for (i = top; i < top + height; i++) {
						 myscreen.settext(i, left, spaces, 0, width);
				 }
		 }

		 private void calculate_font()
		 {
				 Font basefont;
				 int style = Font.PLAIN;

				 basefont = myscreen.fixedfont;

				 if ((curzstyle & BOLD) != 0)
						 style |= Font.BOLD;
				 if ((curzstyle & ITALIC) != 0)
						 style |= Font.ITALIC;

				 curfont = new Font(basefont.getName(), style, basefont.getSize());
		 }

		 public void set_color(int foreground, int background) {
//			System.err.println("fg = " + foreground + ", bg = " + background);
			flush();
			if (foreground != ZColor.Z_CURRENT)
				 zforeground = foreground;
			if (background != ZColor.Z_CURRENT)
				 zbackground = background;
		 }

		 public void set_text_style(int style) {
				 set_text_style(style, false);
		 }

		 protected void set_text_style(int style, boolean immediate) {
				 char thecode[] = new char[1];

				 if (immediate || !buffer) {
						 if (style == ROMAN)
								 curzstyle = ROMAN;
						 else
								 curzstyle |= style;
						 calculate_font();
				 }
				 else {
						 thecode[0] = (char)(style | BUF_ROMAN);
						 bufferchars(thecode);
				 }
		 }

		 private boolean is_control(char ch)
		 {
				 return (ch >= FIRST_STYLE);
		 }
														
		 private void parse_control(char control) {
				 if ((control >= FIRST_STYLE) && (control <= LAST_STYLE)) {
						 set_text_style(control & ~BUF_ROMAN, true);
				 }
		 }

		 public void drawchars(char chars[], int offset, int length) {
				 int firstchar;
				 int runlength;
				 int i;
				 char control = 0;

//				 System.err.println("offset length top left cy cx " + offset + " " +
//												  length + " " + top + " " + left + " " +
//												  cursory + " " + cursorx);
				 if (length != 0) {
						 if (myscreen.zforeground != zforeground)
								 myscreen.setZForeground(zforeground);
						 if (myscreen.zbackground != zbackground)
								 myscreen.setZBackground(zbackground);

						 if (scroll && (cursorx == 0))
								 check_for_more();
						
						 firstchar = offset;
						 runlength = 0;
						 while (firstchar < (offset + length)) {
								 for (i = (firstchar - offset); i < length; i++, runlength++) {
										 if (is_control(chars[offset + i])) {
												 control = chars[offset + i];
												 break;
										 }
								 }
								 myscreen.settext(top + cursory, left + cursorx, chars, 
																  firstchar, runlength,
																  (curzstyle & REVERSE) == REVERSE,
																  curfont);
								 parse_control(control);
								 control = 0;
								 firstchar += runlength + 1;
								 cursorx += runlength;
								 runlength = 0;
						 }
//				 System.err.print(new String(chars, offset, length));
				 }
		 }

		 public void drawstring(String text)
		 {
				 drawchars(text.toCharArray(), 0, text.length());
		 }

		 public int charsWidth(char line[], int offset, int length) {
				 return myscreen.charwidth() * length;
		 }
 }

 class ZCursor {
		Color cursorcolor, bgcolor;
		boolean shown;
		int t,l,w,h;
		Component parent;
		
		ZCursor(Color cursorcolor, Color bgcolor, Component parent) {
				shown = false;
				this.cursorcolor = cursorcolor;
				this.bgcolor = bgcolor;
				this.parent = parent;
		}

		ZCursor(Component parent) {
				this(Color.green, Color.yellow, parent);
		}

		ZCursor() {
				this(Color.green, Color.yellow, null);
		}

		synchronized void show() {
			Graphics g;
			
			if (!shown) {
				shown = true;
				if (parent != null) {
					g = parent.getGraphics();
					if (g != null) {
						g.setColor(cursorcolor);
						g.fillRect(l,t,w,h);
					}
				}
			}
		}

		synchronized void hide() {
			if (shown) {
				shown = false;
				if (parent != null) {
					parent.repaint(l,t,w,h);
//					g.setColor(bgcolor);
//					g.fillRect(l,t,w,h);
				}
			}
		}

		synchronized void redraw(Graphics g) {
			if (shown) {
				g.setColor(cursorcolor);
				g.fillRect(l,t,w,h);
			}
		}

		synchronized void move(int l, int t) {
			boolean wasshown = shown;

			if (wasshown)
					hide();
			this.l = l;
			this.t = t;
			if (wasshown)
					show();
		}

		synchronized void size(int w, int h) {
				boolean wasshown = shown;

				if (wasshown)
						hide();
				this.w = w;
				this.h = h;
				if (wasshown)
						show();
		}

//		synchronized void setGraphics(Graphics g) {
//				boolean wasshown = shown;
//
//				this.g = g;
//				g.setColor(cursorcolor);
//				g.setXORMode(bgcolor);
//				if (wasshown)
//						show();
//		}

		synchronized void setcolors(Color cursorcolor, Color bgcolor) {
				boolean wasshown = shown;

				if (wasshown)
					hide();
				this.cursorcolor = cursorcolor;
				this.bgcolor = bgcolor;
				if (wasshown)
					show();
		}
 }

 class ZStatus extends Panel {
		 boolean timegame;
		 boolean initialized;
		 boolean chronograph;
		 String location;
		 int score;
		 int turns;
		 int hours;
		 int minutes;
		 Label Right;
		 Label Left;
		
		 public ZStatus() {
				 setLayout(new BorderLayout());
				 Right = new Label();
				 add("East", Right);
				 Left = new Label();
				 add("West", Left);
				 chronograph = false;
		 }

/*
		 public boolean gotFocus(Event evt, Object what)
		 {
				 System.err.println("ZStatus got focus");
				 return false;
		 }
		
		 public boolean lostFocus(Event evt, Object what)
		 {
				 System.err.println("ZStatus lost focus");
				 return false;
		 }
*/

		 public void update_score_line(String location, int score, int turns) {
				 this.timegame = false;
				 this.location = location;
				 this.score = score;
				 this.turns = turns;
				 Left.setText(location);
				 Right.setText(score + "/" + turns);
				 layout();
				 repaint();
		 }
		
		 public void update_time_line(String location, int hours, int minutes) {
				 String meridiem;

				 this.timegame = true;
				 this.location = location;
				 this.hours = hours;
				 this.minutes = minutes;
				 Left.setText(location);
				 if (chronograph) {
						 Right.setText(hours + ":" + minutes);
				 }
				 else {
						 if (hours < 12)
								 meridiem = "AM";
						 else
								 meridiem = "PM";
						 hours %= 12;
						 if (hours == 0)
								 hours = 12;
						 Right.setText(hours + ":" + minutes + meridiem);
				 }
				 layout();
				 repaint();
		 }
		
		 public Dimension minimumSize() {
				 return new Dimension(100,10);
		 }
		
		 public Dimension preferredSize() {
				 return new Dimension(500,20);
		 }
 }


 class ZScreen extends Canvas {
		int lines;
		int chars; /* in fixed font */
		Font fixedfont;
		FontMetrics fixedmetrics;
		Font variablefont;
		Font graphicsfont;
		SyncVector inputcodes;
		Vector bufferedcodes;
		boolean bufferdone;
		ZWindow inputwindow;
		ZCursor inputcursor;
		int zforeground = ZColor.Z_BLACK;
		int zbackground = ZColor.Z_WHITE;
		Image backing_store;
		Graphics g_store;
		Color zbcolor;
		boolean hasscrolled = false;
		final static char accent_table[] = {
			 	'\u00e4',			/* a-umlaut */
			 	'\u00f6',			/* o-umlaut */
			 	'\u00fc',			/* u-umlaut */
			 	'\u00c4',			/* A-umlaut */
			 	'\u00d6',			/* O-umlaut */
			 	'\u00dc',			/* U-umlaut */
			 	'\u00df',			/* sz-ligature */
			 	'\u00bb',			/* right-pointing quote */
			 	'\u00ab',			/* left-pointing quote */
			 	'\u00eb',			/* e-umlaut */
			 	'\u00ef',			/* i-umlaut */
			 	'\u00ff',			/* y-umlaut */
			 	'\u00cb',			/* E-umlaut */
			 	'\u00cf',			/* I-umlaut */
			 	'\u00e1',			/* a-acute */
			 	'\u00e9',			/* e-acute */
			 	'\u00ed',			/* i-acute */
			 	'\u00f3',			/* o-acute */
			 	'\u00fa',			/* u-acute */
			 	'\u00fd',			/* y-acute */
			 	'\u00c1',			/* A-acute */
			 	'\u00c9',			/* E-acute */
			 	'\u00cd',			/* I-acute */
			 	'\u00d3',			/* O-acute */
			 	'\u00da',			/* U-acute */
			 	'\u00dd',			/* Y-acute */
			 	'\u00e0',			/* a-grave */
			 	'\u00e8',			/* e-grave */
			 	'\u00ec',			/* i-grave */
			 	'\u00f2',			/* o-grave */
			 	'\u00f9',			/* u-grave */
			 	'\u00c0',			/* A-grave */
			 	'\u00c8',			/* E-grave */
			 	'\u00cc',			/* I-grave */
			 	'\u00d2',			/* O-grave */
			 	'\u00d9',			/* U-grave */
			 	'\u00e2',			/* a-circumflex */
			 	'\u00ea',			/* e-circumflex */
			 	'\u00ee',			/* i-circumflex */
			 	'\u00f4',			/* o-circumflex */
			 	'\u00fb',			/* u-circumflex */
			 	'\u00c2',			/* A-circumflex */
			 	'\u00ca',			/* E-circumflex */
			 	'\u00ce',			/* I-circumflex */
			 	'\u00d4',			/* O-circumflex */
			 	'\u00da',			/* U-circumflex */
			 	'\u00e5',			/* a-ring */
			 	'\u00c5',			/* A-ring */
			 	'\u00f8',			/* o-slash */
			 	'\u00d8',			/* O-slash */
			 	'\u00e3',			/* a-tilde */
			 	'\u00f1',			/* n-tilde */
			 	'\u00f5',			/* o-tilde */
			 	'\u00c3',			/* A-tilde */
			 	'\u00d1',			/* N-tilde */
			 	'\u00d5',			/* O-tilde */
			 	'\u00e6',			/* ae-ligature */
			 	'\u00c6',			/* AE-ligature */
			 	'\u00e7',			/* c-cedilla */
			 	'\u00c7',			/* C-cedilla */
			 	'\u00fe',			/* Icelandic thorn */
			 	'\u00f0',			/* Icelandic eth */
			 	'\u00de',			/* Icelandic Thorn */
			 	'\u00d0',			/* Icelandic Eth */
			 	'\u00a3',			/* UK pound symbol */
			 	'\u0153',			/* oe ligature */
			 	'\u0152',			/* OE ligature */
			 	'\u00a1',			/* inverse-! */
			 	'\u00bf',			/* inverse-? */
	     };


		public ZScreen() {
				Dimension mysize = size();
				
				fixedfont = new Font("Courier", Font.PLAIN, 12);
				fixedmetrics = getFontMetrics(fixedfont);
				chars = mysize.width / fixedmetrics.charWidth(' ');
				lines = mysize.height / fixedmetrics.getHeight();

				inputcodes = new SyncVector();
				bufferedcodes = new Vector();
				inputcursor = new ZCursor(this);
				setForeground(ZColor.getcolor(zforeground));
				setBackground(ZColor.getcolor(zbackground));
		}
		
		protected boolean isterminator(int key) {
				return ((key == 10) || (key == 13));
		}

/*		public boolean mouseDown(Event evt, int x, int y) {
//				System.err.print("mouseDown ");
				if (inside(x,y)) {
						requestFocus();
//						System.err.println("inside");
						return true;
				}
				else {
						System.err.println("outside");
						return false;
				}
		}
*/

		public boolean mouseEnter(Event evt, int x, int y)
		{
				requestFocus();
				return true; /* it's MY event */
		}

/*
		public boolean gotFocus(Event evt, Object what)
		{
				System.err.println("ZScreen got focus");
				return false;
		}
		
		public boolean lostFocus(Event evt, Object what)
		{
				System.err.println("ZScreen lost focus");
				return false;
		}
*/

		static char zascii_to_unicode(short zascii) {
				if ((zascii >= 32) && (zascii <= 126)) /* normal ascii */
					return (char)zascii;
				else if ((zascii >= 155) && (zascii <= 251)) {
					if ((zascii - 155) < accent_table.length) {
						return accent_table[zascii - 155];
					}
					else
						return '?';
				}
				else if ((zascii == 0) || (zascii >= 256)) {
					return '?';
				}
				else {
					System.err.println("Illegal character code: " + zascii);
					return '?';
				}
		}

		static short unicode_to_zascii(char unicode) throws NoSuchKeyException {
				short i;
				
				if (unicode == '\n')
					return 13;
				if (unicode == '\b')
					return 127;
				else if (((int)unicode < 0x20) &&
					(unicode != '\uu000d') &&
					(unicode != '\uu001b'))
					throw new NoSuchKeyException("Illegal character input: " + (short)unicode);
				else if ((int)unicode < 0x80) /* normal ascii, including DELETE */
					return (short)unicode;
				else {
					for (i = 0; i < accent_table.length; i++) {
						if (accent_table[i] == unicode)
							return (short)(155 + i);
					}
					throw new NoSuchKeyException("Illegal character input: " + (short)unicode);
				}
		}

		static short fkey_to_zascii(int fkey) throws NoSuchKeyException {
				switch (fkey) {
					case Event.UP:		return 129;
					case Event.DOWN:	return 130;
					case Event.LEFT:	return 131;
					case Event.RIGHT:	return 132;
					case Event.F1:		return 133;
					case Event.F2:		return 134;
					case Event.F3:		return 135;
					case Event.F4:		return 136;
					case Event.F5:		return 137;
					case Event.F6:		return 138;
					case Event.F7:		return 139;
					case Event.F8:		return 140;
					case Event.F9:		return 141;
					case Event.F10:		return 142;
					case Event.F11:		return 143;
					case Event.F12:		return 144;
					default:
						throw new NoSuchKeyException("Illegal function key " + fkey);
				}
		}
		
		public boolean keyDown(Event e, int key) {

				short code;
				
				/* TODO: e, key to code */
				try {
					if (e.id == Event.KEY_PRESS)
						code = unicode_to_zascii((char)key);
					else /* if (e.action == Event.KEY_ACTION) */
						code = fkey_to_zascii(key);
					inputcodes.syncAddElement(new Integer(code));
				}
				catch (NoSuchKeyException excpt) {
					System.err.println(excpt);
				}
				return true;
		}
		
		public void set_input_window(ZWindow thewindow)
		{
				inputwindow = thewindow;
		}

		public short read_code() {
				Integer thecode = null;
				
				while (thecode == null) {
						thecode = (Integer)inputcodes.syncPopFirstElement();
				}
				return (short)thecode.intValue();
		}
		
		public short read_buffered_code() { /* should really be synched */
						   Integer thecode;
				int incode;
				short result;
				int cw, ch;

				inputwindow.flush();
				cw = fixedmetrics.charWidth(' ');
				ch = fixedmetrics.getHeight();

//				inputcursor.setGraphics(getGraphics());
//				inputcursor.setcolors(getForeground(), getBackground());
				inputcursor.setcolors(getForeground(), zbcolor);
				inputcursor.size(cw, ch);

				while (!bufferdone) {
						inputwindow.flush();
						inputcursor.move((inputwindow.getLeft() + inputwindow.cursorx) * cw,
														 (inputwindow.getTop() + inputwindow.cursory) * ch);
						inputcursor.show();
						Toolkit.getDefaultToolkit().sync();
						incode = read_code();
//						inputcursor.setGraphics(getGraphics());
						inputcursor.hide();
						if ((incode == 8) || (incode == 127)) {
								try {
										thecode = (Integer)bufferedcodes.lastElement();
										bufferedcodes.removeElementAt(bufferedcodes.size() - 1);
										inputwindow.flush();
										inputwindow.movecursor(inputwindow.cursorx - 1, 
																				   inputwindow.cursory);
										inputwindow.printzascii((short)' ');
										inputwindow.flush();
										inputwindow.movecursor(inputwindow.cursorx - 1, 
																				   inputwindow.cursory);
								}
								catch (NoSuchElementException booga) {
										/* ignore */
								}
						}
						else {
								if (isterminator(incode)) {
										bufferdone = true;
										if ((incode==10) || (incode == 13))
												inputwindow.newline();
								}
								else {
										inputwindow.printzascii((short)incode);
										inputwindow.flush();
								}
								bufferedcodes.addElement(new Integer(incode));
						}
				}
				thecode = (Integer)bufferedcodes.firstElement();
				bufferedcodes.removeElementAt(0);
				if (bufferedcodes.isEmpty()) {
						bufferdone = false;
				}
				return (short)(thecode.intValue());
		}

		public int getlines() {
				return lines;
		}
		
		public int getchars() {
				return chars;
		}

		public synchronized void reshape(int x, int y, int width, int height) {
				int lines;
				int i;
				char newbuffer[][];

				if ((width >= 0) && (height >= 0)) {
						lines = height / fixedmetrics.getHeight();
						chars = width / fixedmetrics.charWidth(' ');
						backing_store = createImage(width, height);
						g_store = backing_store.getGraphics();
						g_store.setColor(zbcolor);
						g_store.fillRect(0, 0, width, height);
						this.lines = lines;
				}
				super.reshape(x, y, width, height);
		}
		
		public int charwidth() { /* character width of the fixed font */
				return fixedmetrics.charWidth(' ');
		}

		public synchronized void settext(int y, int x, char newtext[],
																		 int offset, int length)
		{
				settext(y,x,newtext,offset,length,false, fixedfont);
		}

		public synchronized void settext(int y, int x, char newtext[],
																		 int offset, int length, boolean reverse,
																		 Font textfont) {
				int i;
				char newbuffer[];
				Graphics g;

				try {
						g_store.setFont(textfont);
						drawtext(g_store, y, x, newtext, offset, length, reverse);
						if (!hasscrolled) {
								g = getGraphics();
								g.setFont(textfont);
								drawtext(g, y, x, newtext, offset, length, reverse);
						}
				}
				catch (NullPointerException booga) {
						System.err.println("No graphics in settext");
				}
		}

		protected synchronized void drawtext(Graphics g, int y, int x, char newtext[], int offset, int length, boolean reverse) {
				int tw, th;
				int tx, ty;
				
				tw = length * fixedmetrics.charWidth(' ');
				th = fixedmetrics.getHeight();
				tx = x * fixedmetrics.charWidth(' ');
				ty = th * (y + 1) - fixedmetrics.getDescent();
				if (reverse) {
						g.setColor(getForeground());
						g.fillRect(tx, th * y, tw, th);
//						g.setColor(getBackground());
						g.setColor(zbcolor);
				}
				else {
//						g.setColor(getBackground());
						g.setColor(zbcolor);
						g.fillRect(tx, th * y, tw, th);
						g.setColor(getForeground());
				}
				g.drawChars(newtext, offset, length, tx, ty);
				g.setColor(getForeground());
		}
		
		public synchronized void scrollLines(int top, int height, int lines) {
				int j;
				Graphics g;
				int texttop;

//				System.err.println("scrollLines top height lines "
//												   + top + " "
//												   + height + " "
//												   + lines + " ");
				try {
						texttop = top * fixedmetrics.getHeight();
						g_store.copyArea(0, texttop + lines * fixedmetrics.getHeight(),
														 size().width, (height - lines) * fixedmetrics.getHeight(),
														 0, -lines * fixedmetrics.getHeight());
//						g_store.setColor(getBackground());
						g_store.setColor(zbcolor);
						g_store.fillRect(0, texttop + ((height-1) * fixedmetrics.getHeight()),
														 size().width, fixedmetrics.getHeight());
//						Toolkit.getDefaultToolkit().sync();
				}
				catch (NullPointerException booga) {
						System.err.println("No graphics in scrollLines");
				}
				repaint();
				hasscrolled = true;
		}
		
		public synchronized void paint(Graphics g) {
				int y;
				int ypixels;

//				super.paint(g);
//				System.err.println("paint");
				g.drawImage(backing_store, 0, 0, null);
				inputcursor.redraw(g);
				hasscrolled = false;
		}

//		public void repaint(long tm, int x, int y, int width, int height) {
//				System.err.println("repaint: " + tm + " " + x + " " + y + " " + width + " " + height);
//				super.repaint(tm,x,y,width,height);
//		}

		public void update(Graphics g) {
//				System.err.println("update");
				g.setColor(getForeground());
				paint(g);
		}

		public void clear() {
				Dimension mysize = size();

				try {
//						g_store.setColor(getBackground());
						g_store.setColor(zbcolor);
						g_store.fillRect(0, 0,
														 mysize.width, mysize.height);
				}
				catch (NullPointerException booga) {
						System.err.println("No graphics in clear");
				}
				repaint();
		}

		public int getZForeground() 
		{
				return zforeground;
		}
		
		public int getZBackground() 
		{
				return zbackground;
		}
		
		public void setZForeground(int zcolor)
		{
				zforeground = zcolor;
				setForeground(ZColor.getcolor(zcolor));
		}
		
		public void setZBackground(int zcolor)
		{
				zbackground = zcolor;
//				setBackground(ZColor.getcolor(zcolor)); Yucky side effects
				zbcolor = ZColor.getcolor(zcolor);
		}
		
		public Frame getFrame() {
			Component cursor;
			
			cursor = this;
			while (!(cursor instanceof Frame))
				cursor = cursor.getParent();
				
			return (Frame)cursor;
		}
		
		public Dimension minimumSize() {
				return new Dimension(100,100);
		}

		public Dimension preferredSize() {
				return new Dimension(500,500);
		}
}
