/* Zplet, a Z-Machine interpreter in Java */
/* Copyright 1996 Matthew T. Russotto */
/* All Rights Reserved */
/* You may freely copy, modify, and make derivative works from this code
   provided my copyright notice is retained and any modifications are
   clearly marked */

import java.util.*;
import java.awt.*;
import java.net.*;
import java.io.*;
import java.applet.Applet;

public class Zplet extends java.applet.Applet {
		ZScreen screen;
		ZStatus status_line;
		ZMachine zm;

		boolean failed = false;
		
		public void init() {
			String statusfg, statusbg, mainfg, mainbg;
			Component parent;
			
			statusfg = getParameter("StatusForeground");
			statusbg = getParameter("StatusBackground");
			mainfg = getParameter("Foreground");
			mainbg = getParameter("Background");
			setLayout(new BorderLayout());
			screen = new ZScreen();
			status_line = new ZStatus();
			status_line.setForeground(Color.black);
			status_line.setBackground(Color.white);
			screen.setZForeground(ZColor.Z_WHITE);
			screen.setZBackground(ZColor.Z_BLACK);
			if (statusfg != null)
				status_line.setForeground(ZColor.getcolor(statusfg));

			if (statusbg != null)
				status_line.setBackground(ZColor.getcolor(statusbg));

			if (mainfg != null)
				screen.setZForeground(ZColor.getcolornumber(mainfg));
			if (mainbg != null)
				screen.setZBackground(ZColor.getcolornumber(mainbg));

			add("North", status_line);
			add("Center", screen);
			
		}
		
		void startzm() {
			URL myzfile;
			InputStream myzstream;
			String zcodefile;

			byte zmemimage[];

			zmemimage = null;
			try {
				zcodefile = getParameter("StoryFile");
				myzfile = new URL(getCodeBase(),zcodefile);
//				System.err.println(myzfile);
				myzstream = myzfile.openStream();
				zmemimage = suckstream(myzstream);
			}
			catch (MalformedURLException booga) {
				add ("North", new Label("Malformed URL"));
				failed = true;
			}
			catch (IOException booga) {
				add ("North", new Label("I/O Error"));
				/* don't set failed, may want to retru */
			}
			if (zmemimage != null) {
				switch (zmemimage[0]) {
				case 3:
					zm = new ZMachine3(screen, status_line, zmemimage);
					break;
				case 5:
					remove(status_line);
					zm = new ZMachine5(screen, zmemimage);
					break;
				case 8:
					remove(status_line);
					zm = new ZMachine8(screen, zmemimage);
					break;
				default:
					add ("North", new Label("Not a valid V3,V5, or V8 story file"));
				}
				if (zm != null)
					zm.start();
			}
		}

		byte [] suckstream(InputStream mystream) throws IOException {
			byte buffer[];
			byte oldbuffer[];
			int currentbytes = 0;
			int bytesleft;
			int got;
			int buffersize = 2048;


			buffer = new byte[buffersize];
			bytesleft = buffersize;
			got = 0;
			while (got != -1) {
				bytesleft -= got;					
				currentbytes += got;
				if (bytesleft == 0) {
					oldbuffer = buffer;
					buffer = new byte[buffersize + currentbytes];
					System.arraycopy(oldbuffer, 0, buffer, 0, currentbytes);
					oldbuffer = null;
					bytesleft = buffersize;
				}
				got = mystream.read(buffer, currentbytes, bytesleft);
			}
			if (buffer.length != currentbytes) {
				oldbuffer = buffer;
				buffer = new byte[currentbytes];
				System.arraycopy(oldbuffer, 0, buffer, 0, currentbytes);
			}
			return buffer;
		}

/*		public boolean gotFocus(Event evt, Object what)
		{
			System.err.println("ZPlet got focus");
			return false;
		}
		
		public boolean lostFocus(Event evt, Object what)
		{
			System.err.println("ZPlet lost focus");
			return false;
		}
*/
		public void start() {
			if (!failed && ((zm==null) || !zm.isAlive())) {
				startzm();
			}
				
		}

		public void destroy() {
			zm.stop();
			zm = null;
			remove(screen);
			screen = null;
		}
}
